import sys
import threading
from unittest.mock import patch
from typing import Callable

from pydantic import BaseModel, Field

from versionhq.agent.model import Agent, LLM
from versionhq.task.model import Task, ResponseField, TaskOutput, TaskExecutionType
from versionhq.tool.model import Tool, ToolSet
from versionhq.tool.decorator import tool

sys.setrecursionlimit(2097152)
threading.stack_size(134217728)



def test_async_execute_task():
    task = Task(description="Return string: 'test'", type=TaskExecutionType.ASYNC)

    with patch.object(Agent, "execute_task", return_value="test") as execute:
        res = task.execute()
        assert res.raw == "test"
        execute.assert_called_once_with(task=task, context=None, task_tools=list())


def test_sync_execute_with_task_context():
    """
    Use case = One agent handling multiple tasks sequentially using context set in the main task.
    """
    sub_task = Task(
        description="return the output following the given prompt.",
        response_fields=[
            ResponseField(title="subtask_result", data_type=str, required=True),
        ]
    )
    main_task = Task(
        description="return the output following the given prompt.",
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=str, required=True),
        ],
        context=[sub_task,]
    )
    res = main_task.execute()

    assert isinstance(res, TaskOutput)
    assert res.task_id is main_task.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is None
    assert sub_task.output is not None
    assert sub_task.output.json_dict is not None
    assert "subtask_result" in main_task.prompt()


def test_sync_execute_task_with_prompt_context():
    """
    Use case:
    - One agent handling multiple tasks sequentially using context set in the main task.
    - On top of that, the agent receives context when they execute the task.
    """
    class Outcome(BaseModel):
        test1: int = Field(default=None)
        test2: str = Field(default=None)

    sub_task = Task(
        description="return the output following the given prompt.",
        response_fields=[ResponseField(title="result", data_type=str, required=True),]
    )
    main_task = Task(
        description="return the output following the given prompt.",
        pydantic_output=Outcome,
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=str, required=True),
        ],
        context=[sub_task]
    )
    res = main_task.execute(context="plan a Black Friday campaign.")

    assert isinstance(res, TaskOutput) and res.task_id is main_task.id
    assert res.raw and isinstance(res.raw, str)
    assert res.json_dict and isinstance(res.json_dict, dict)
    assert res.pydantic.test1 == res.json_dict["test1"] and res.pydantic.test2 == res.json_dict["test2"]
    assert sub_task.output is not None
    assert "result" in main_task.prompt()
    assert main_task.prompt_context == "plan a Black Friday campaign."
    assert "plan a Black Friday campaign." in main_task.prompt()


def test_callback():
    """
    See if the callback function is executed well with kwargs.
    """

    def callback_func(condition: str, test1: str):
        # task_id = str(id) if id else None
        return f"Result: {test1}, condition added: {condition}"

    task = Task(
        description="return the output following the given prompt.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
        ],
        callback=callback_func,
        callback_kwargs=dict(condition="demo for pytest")
    )
    res = task.execute()

    assert res and isinstance(res, TaskOutput)
    assert res.task_id is task.id
    assert "demo for pytest" in res.callback_output


def test_delegate():
    agent = Agent(role="demo agent 6", goal="My amazing goals", maxit=1, max_tokens=3000)
    task = Task(
        description="return the output following the given prompt.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
        ],
        allow_delegation=True
    )
    task.execute()

    assert task.output is not None
    assert "vhq-Delegated-Agent" in task.processed_agents
    assert task.delegations != 0


# def test_conditional_task():
#     task = Task(
#         description="erturn the output following the given prompt.",
#         response_fields=[ResponseField(title="test1", data_type=str, required=True),],
#     )
#     res = task.execute_sync(agent=base_agent)

#     conditional_task = ConditionalTask(
#         description="return the output following the given prompt.",
#         response_fields=[ResponseField(title="test1", data_type=str, required=True),],
#         condition=lambda x: bool("zzz" in task.output.raw)
#     )
#     should_execute = conditional_task.should_execute(context=res)

#     assert res.raw is not None
#     assert should_execute is False

#     conditional_res = conditional_task._handle_conditional_task(task_outputs=[res,], task_index=1, was_replayed=False)
#     if not should_execute:
#         assert conditional_res is None
#     else:
#         assert conditional_res.task_id is conditional_task.id


def test_store_task_log():
    task = Task(
        description="return the output following the given prompt.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    assert task._task_output_handler.load() is not None


def test_task_with_agent_tools():
    simple_tool = Tool(name="simple tool", func=lambda x: "simple func")
    agent = Agent(role="demo", goal="execute tools", tools=[simple_tool,], maxit=1, max_tokens=3000)
    task = Task(description="execute tool", can_use_agent_tools=True, tool_res_as_final=True)
    res = task.execute(agent=agent)
    assert res.tool_output == "simple func"

    def empty_func():
        return "empty func"

    func_tool = Tool(name="func tool", func=empty_func)
    agent.tools = [func_tool]
    res = task.execute(agent=agent)
    assert res.tool_output == "empty func"


    def demo_func(message: str) -> str:
        return message + "_demo"

    class CustomTool(Tool):
        name: str = "custom tool"

    custom_tool = CustomTool(func=demo_func)
    agent.tools = [custom_tool]
    res = task.execute(agent=agent)
    assert "_demo" in res.tool_output


def test_task_with_tools():

    def random_func(message: str) -> str:
        return message + "_demo"

    tool = Tool(name="tool", func=random_func)
    tool_set = ToolSet(tool=tool, kwargs=dict(message="empty func"))
    task = Task(description="execute the given tools", tools=[tool_set,], tool_res_as_final=True)
    res = task.execute()
    assert res.tool_output == "empty func_demo"

    class CustomTool(Tool):
        name: str = "custom tool"
        func: Callable = None

    custom_tool = CustomTool(func=random_func)
    task.tools = [custom_tool]
    res = task.execute()
    assert "_demo" in res.tool_output

    task.tools = [custom_tool]
    res = task.execute()
    assert res.tool_output is not None



def test_task_without_response_format():
    task = Task(description="return a simple output with any random values.")
    res = task.execute()

    assert res and isinstance(res, TaskOutput)
    assert res.json_dict and isinstance(res.json_dict, dict)
    assert res.pydantic is None


def test_build_agent_without_developer_prompt():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        backstory="You are competitive analysts who have abundand knowledge in marketing, product management.",
        use_developer_prompt=False
    )
    task = Task(description="return a simple output with any random values.")
    res = task.execute(agent=agent)

    assert res and isinstance(res, TaskOutput)
    assert res.json_dict and isinstance(res.json_dict, dict)
    assert res.pydantic is None


def test_callback_with_custom_output():
    class CustomOutput(BaseModel):
        test1: str
        test2: list[str]

    def dummy_func(message: str, test1: str, test2: list[str]) -> str:
        return f"""{message}: {test1}, {", ".join(test2)}"""

    task = Task(
        description="Amazing task",
        pydantic_output=CustomOutput,
        callback=dummy_func,
        callback_kwargs=dict(message="Hi! Here is the result: ")
    )
    res = task.execute(context="amazing context to consider.")

    assert res.task_id == task.id
    assert res.pydantic.test1 and res.pydantic.test2
    assert "Hi! Here is the result: " in res.callback_output and res.pydantic.test1 in res.callback_output and ", ".join(res.pydantic.test2) in res.callback_output


def test_task_with_agent_callback():
    import litellm

    def dummy_func(*args, **kwargs) -> str:
        return "Demo func"

    agent = Agent(role="demo", goal="amazing project goal", maxit=1, max_tokens=3000, callbacks=[dummy_func,])
    task = Task(description="Amazing task")
    res = task.execute(agent=agent)

    assert res.raw and res.task_id == task.id
    assert litellm.callbacks == [dummy_func]


def test_rpm():
    agent = Agent(role="demo", goal="use the given tools", max_tokens=3000, max_rpm=3)
    assert agent._rpm_controller and agent._rpm_controller.max_rpm == 3

    a = "hello"
    tool = Tool(func=lambda x: a + x)
    tool_set = ToolSet(tool=tool, kwargs={ "x": "_demo" })
    task = Task(description="Summarize overview of the given tool in sentence, then execute the tool.", tools=[tool_set,])
    res = task.execute(agent=agent)

    assert "hello_demo" in res.raw if res.raw else res
    assert agent._rpm_controller._current_rpm < agent._rpm_controller.max_rpm


def test_maxit():
    @tool
    def demo() -> str:
        """Get the final answer but don't give it yet, just re-use this
        tool non-stop."""
        return "demo"

    agent = Agent(role="demo", goal="amazing demo", maxit=2)
    task = Task(description="Summarize overview of the given tool in sentences.", tools=[demo,])

    with patch.object(LLM, "call", wraps=agent.llm.call) as mock:
        task.execute(agent=agent)
        assert mock.call_count <= 2


def test_evaluation():
    """
    See if the output will be evaluated accurately - when the task was given eval criteria
    """
    from versionhq.task.evaluate import Evaluation, EvaluationItem
    from versionhq.agent.inhouse_agents import vhq_task_evaluator

    task = Task(
        description="Research a topic to teach a kid aged 6 about math.",
        should_evaluate=True,
        eval_criteria=["Uniquness of the topic researched", "Fit to the target audience",],
    )
    res = task.execute()

    assert res.evaluation and isinstance(res.evaluation, Evaluation)
    assert [isinstance(item, EvaluationItem) and item.criteria in task.eval_criteria for item in res.evaluation.items]
    assert res.evaluation.latency and res.evaluation.tokens and res.evaluation.eval_by == vhq_task_evaluator
    assert res.evaluation.aggregate_score is not None and res.evaluation.suggestion_summary



"""
Test use cases on index.md and quickstart.md
"""

def test_doc_index_a():
    import versionhq as vhq

    network = vhq.form_agent_network(
        task="YOUR AMAZING TASK OVERVIEW",
        expected_outcome="YOUR OUTCOME EXPECTATION",
    )
    res = network.launch()
    assert res.raw is not None


def test_doc_index_b():
    import versionhq as vhq
    from pydantic import BaseModel

    class CustomOutput(BaseModel):
        test1: str
        test2: list[str]

    def dummy_func(message: str, test1: str, test2: list[str]) -> str:
        return f"""{message}: {test1}, {", ".join(test2)}"""


    agent = vhq.Agent(role="demo", goal="amazing project goal", maxit=1)

    task = vhq.Task(
        description="Amazing task",
        pydantic_output=CustomOutput,
        callback=dummy_func,
        callback_kwargs=dict(message="Hi! Here is the result: ")
    )

    res = task.execute(agent=agent, context="amazing context to consider.")
    assert "Hi! Here is the result:" in res.callback_output
    assert [getattr(res.pydantic, k) for k, v in CustomOutput.model_fields.items()]



"""Test code snippets on docs/core/task.md"""


def test_docs_core_task_a():
    import versionhq as vhq

    task = vhq.Task(description="MY AMAZING TASK")

    import uuid
    assert uuid.UUID(str(task.id), version=4)


def test_docs_core_task_b():
    import versionhq as vhq

    task = vhq.Task(description="MY AMAZING TASK")
    res = task.execute()

    assert isinstance(res, vhq.TaskOutput)
    assert res.raw and res.json
    assert task.processed_agents is not None



def test_docs_core_task_c():
    from pydantic import BaseModel
    from typing import Any

    class Demo(BaseModel):
        """
        A demo pydantic class to validate the outcome with various nested data types.
        """
        demo_1: int
        demo_2: float
        demo_3: str
        demo_4: bool
        demo_5: list[str]
        demo_6: dict[str, Any]
        demo_nest_1: list[dict[str, Any]] # 1 layer of nested child is ok.
        demo_nest_2: list[list[str]]
        demo_nest_3: dict[str, list[str]]
        demo_nest_4: dict[str, dict[str, Any]]
        # error_1: list[list[dict[str, list[str]]]] # <- Error due to 2+ layers of nested children.
        # error_2: InstanceOf[AnotherPydanticClass] # <- Error due to non-typing annotation.

    # 2. Define a task
    import versionhq as vhq
    task = vhq.Task(
        description="generate random output that strictly follows the given format",
        pydantic_output=Demo,
    )

    # 3. Execute
    res = task.execute()

    assert isinstance(res, vhq.TaskOutput) and res.task_id is task.id
    assert isinstance(res.raw, str) and isinstance(res.json_dict, dict)
    assert [
        getattr(res.pydantic, k) and v.annotation == Demo.model_fields[k].annotation
        for k, v in res.pydantic.model_fields.items()
    ]



def test_docs_core_task_d():
    import versionhq as vhq

    # 1. Define a list of ResponseField objects
    demo_response_fields = [
        vhq.ResponseField(title="demo_1", data_type=int),
        vhq.ResponseField(title="demo_2", data_type=float),
        vhq.ResponseField(title="demo_3", data_type=str),
        vhq.ResponseField(title="demo_4", data_type=bool),
        vhq.ResponseField(title="demo_5", data_type=list, items=str),
        vhq.ResponseField(title="demo_6", data_type=dict, properties=[vhq.ResponseField(title="demo-item", data_type=str)]),
        vhq.ResponseField(title="demo_nest_1", data_type=list, items=dict, properties=([
            vhq.ResponseField(title="nest1", data_type=dict, properties=[vhq.ResponseField(title="nest11", data_type=str)])
        ])),
        vhq.ResponseField(title="demo_nest_2", data_type=list, items=list),
        vhq.ResponseField(title="demo_nest_3", data_type=dict, properties=[
            vhq.ResponseField(title="nest1", data_type=list, items=str)
        ]),
        vhq.ResponseField(title="demo_nest_4", data_type=dict, properties=[
            vhq.ResponseField(title="nest1", data_type=dict, properties=[vhq.ResponseField(title="nest12", data_type=str)])
        ])
    ]

    task = vhq.Task(
        description="Output random values strictly following the data type defined in the given response format.",
        response_fields=demo_response_fields
    )

    res = task.execute()

    assert isinstance(res, vhq.TaskOutput) and res.task_id is task.id
    assert res.raw and res.json and res.pydantic is None
    assert [v and type(v) == task.response_fields[i].data_type for i, (k, v) in enumerate(res.json_dict.items())]



def test_docs_core_task_d():
    import versionhq as vhq
    from pydantic import BaseModel
    from typing import Any

    # 1. Define and execute a sub task with Pydantic output.
    class Sub(BaseModel):
        sub1: list[dict[str, Any]]
        sub2: dict[str, Any]

    sub_task = vhq.Task(
        description="generate random values that strictly follows the given format.",
        pydantic_output=Sub
    )
    sub_res = sub_task.execute()

    # 2. Define a main task, callback function to format the final response.
    class Main(BaseModel):
        main1: list[Any] # <= assume expecting to store Sub object.
        main2: dict[str, Any]

    def format_response(sub, main1, main2) -> Main:
        main1.append(sub)
        main = Main(main1=main1, main2=main2)
        return main

    # 3. Execute
    main_task = vhq.Task(
        description="generate random values that strictly follows the given format.",
        pydantic_output=Main,
        callback=format_response,
        callback_kwargs=dict(sub=Sub(sub1=sub_res.pydantic.sub1, sub2=sub_res.pydantic.sub2)),
    )
    res = main_task.execute(context=sub_res.raw) # [Optional] Adding sub_task's response as context.

    assert [item for item in res.callback_output.main1 if isinstance(item, Sub)]



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field, PrivateAttr


class Printer:
    def print(self, content: str, color: Optional[str] = None):
        if color == "purple":
            self._print_purple(content)
        elif color == "red":
            self._print_red(content)
        elif color == "green":
            self._print_green(content)
        elif color == "blue":
            self._print_blue(content)
        elif color == "yellow":
            self._print_yellow(content)
        else:
            print(content)

    def _print_purple(self, content):
        print("\033[1m\033[95m {}\033[00m".format(content))

    def _print_green(self, content):
        print("\033[1m\033[92m {}\033[00m".format(content))

    def _print_red(self, content):
        print("\033[91m {}\033[00m".format(content))

    def _print_blue(self, content):
        print("\033[1m\033[94m {}\033[00m".format(content))

    def _print_yellow(self, content):
        print("\033[1m\033[93m {}\033[00m".format(content))


class Logger(BaseModel):
    """
    Control CLI messages.
    Color: red = error, yellow = warning, blue = info (from vhq), green = info (from third parties)
    """

    verbose: bool = Field(default=True)
    _printer: Printer = PrivateAttr(default_factory=Printer)

    def log(self, level, message, color="yellow"):
        if self.verbose:
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            self._printer.print(f"\n{timestamp} - versionHQ [{level.upper()}]: {message}", color=color)
